from collections.abc import Iterable
import math

def get_bytes(n: int | Iterable) -> bytes:
    '''Return bytes from an integer or iterable yielding integers in range(0xFF)'''

    if isinstance(n, Iterable):
        return bytes(n)

    return n.to_bytes(math.ceil(n.bit_length() / 8))

def is_hex(string: str) -> bool:
    return all(i in set('0123456789abcdefABCDEF') for i in string)


class VehicleException(Exception):
    '''raised when vehicle responds with error'''

class UnlockException(VehicleException):
    '''raised when PCM unlock fails'''

class DeviceException(Exception):
    '''raised for scantool errors'''


def seedkey(seed: bytes, algorithm) -> bytes:
    key = int.from_bytes(seed)

    for i in range(1, 13, 3): # opcodes are every 3rd value starting at index 1
        high = algorithm[i+1] # next 2 values are operands
        low = algorithm[i+2]

        match algorithm[i]:
            case 0x14:
                # add HHLL
                key += high << 8 | low

            case 0x2A:
                if high > low:
                    key = ~key # 1's complement
                else:
                    key = ~key + 1 # 2's complement

            case 0x4C:
                # rotate left by HH bits
                key = (key << high | key >> (16 - high)) & 0xFFFF

            case 0x6B:
                # rotate right by LL bits
                key = (key >> low | key << (16 - low)) & 0xFFFF

            case 0x7E:
                # swap bytes, if LL>HH add LLHH, else add HHLL
                key = (key & 0xFF) << 8 | (key & 0xFF00) >> 8
                if low > high:
                    key += low << 8 | high
                else:
                    key += high << 8 | low

            case 0x98:
                # subtract HHLL
                key -= high << 8 | low

            case _:
                continue

        key = key & 0xFFFF # truncate to 2 bytes
        
    return key.to_bytes(2)


from enum import Enum, IntEnum, auto

class PcmType(Enum):
    p01 = auto()
    p04 = auto()

    @property
    def seedkey_algorithm(self):
        return SEEDKEY_ALGORITHM[self]

    @classmethod
    def from_osid(cls, osid):
        return OSID_LOOKUP[osid]

class BlockId(IntEnum):
    vin1 = 0x01 # 0x00, first 5 bytes
    vin2 = 0x02
    vin3 = 0x03
    osid = 0x0A

SEEDKEY_ALGORITHM = {
    # seedkey algorithms found in STGTERM.dat from GM software
    PcmType.p01: (36, 20, 82, 1, 126, 56, 151, 42, 190, 56, 152, 212, 40),
    PcmType.p04: (4, 107, 80, 2, 126, 80, 210, 76, 5, 253, 152, 24, 203),
}

OSID = {
    PcmType.p01: {
        12212156,
    },
    PcmType.p04: {
        9384022,
    },
}

OSID_LOOKUP = {key: value for value, keys in OSID.items() for key in keys}


from enum import IntEnum
import re
from .vpw import (
    VpwMessage,
    Priority,
    DataRate,
    PhysicalAddress,
    FunctionalAddress,
    Mode
)
from .seedkey import seedkey
from .exceptions import VehicleException, UnlockException
from .pcm import PcmType, BlockId

import logging
logger = logging.getLogger(__name__)

class Vehicle:
    '''SAE J1979 modes'''

    def __init__(self, device):
        self._device = device

    def get_pid(self, pid: int) -> bytes:
        '''mode $01 - request PID'''
        assert pid in range(0xFF)

        request = VpwMessage(
            Priority.functional0,
            FunctionalAddress.obd_request,
            PhysicalAddress.scantool,
            Mode.get_pid,
            pid
        )

        response = self._device.send_message(request)[0]
        return response.data

    def get_supported_pids(self) -> list[int]:
        supported = []
        pid = 1
        for byte in self.get_pid(0):
            for i in range(8):
                if (byte << i) & 0b10000000:
                    supported.append(pid)
                pid += 1
        
        return supported

    def get_dtc(self) -> list[str]:
        '''mode 0x03'''
        request = VpwMessage(
            Priority.functional0,
            FunctionalAddress.obd_request,
            PhysicalAddress.scantool,
            Mode.get_dtc
        )
        raise NotImplementedError

    def clear_dtc(self):
        '''mode $04'''
        raise NotImplementedError

    def get_pending_dtc(self) -> list[str]:
        '''mode $07'''
        raise NotImplementedError

    def get_freeze_frame(self):
        '''mode $02'''
        raise NotImplementedError

    def get_test_results(self):
        '''mode $06'''
        raise NotImplementedError

class GmVehicle(Vehicle):
    '''SAE J2190 modes for General Motors J1850 VPW vehicles'''

    def __init__(self, device, **kwargs):
        super().__init__(device)
        self.osid = self.get_osid()
        self.pcm_type = kwargs.pop('pcm_type', None)

        if self.pcm_type is None:
            try:
                self.pcm_type = PcmType.from_osid(self.osid)
            except KeyError:
                logger.warning(f'unknown OSID: {self.osid}')
  
    def get_pid(self, pid: int) -> bytes:
        '''mode $22 - request PID'''
        assert pid in range(0xFFFF)

        request = VpwMessage(
            Priority.physical0,
            PhysicalAddress.pcm,
            PhysicalAddress.scantool,
            Mode.get_pid_ext,
            pid.to_bytes(2),
            DataRate.single_response
        )

        response = self._device.send_message(request)[0]
        return response.data

    def define_dpid(self, dpid: int, pid: int, size: int, offset: int):
        '''mode $2C - define diagnostic data packet'''
        assert dpid in range(0xFF)
        assert pid in range(0xFFFF)
        assert offset >= 1

        byte3 = 1 << 6 | offset << 3 | size # See SAE J2190 5.19

        request = VpwMessage(
            Priority.physical0,
            PhysicalAddress.pcm,
            PhysicalAddress.scantool,
            Mode.define_dpid,
            dpid,
            (byte3, *pid.to_bytes(2), 0xFF, 0xFF)
        )

        response = self._device.send_message(request)[0]

        if response.mode == Mode.general_response:
            raise VehicleException('request refused')

    def get_dpids(self, dpids: list[int, ...]) -> list[bytes]:
        '''mode $2A - request diagnostic data packet'''
        assert 1 <= len(dpids) <= 6
        assert all(d in range(0xFF) for d in dpids)

        size = len(dpids)
        assert size == len(set(dpids))

        dpids.extend([dpids[0] for _ in range(6 - size)]) # need 6 dpids

        request = VpwMessage(
            Priority.physical0,
            PhysicalAddress.pcm,
            PhysicalAddress.scantool,
            Mode.get_dpid,
            DataRate.single_response,
            dpids
        )
        # must receive all 6 responses before sending anything else
        responses = self._device.send_message(request, 6)

        data = []
        for response in responses[:size]: # ignore duplicates
            if response.mode == Mode.general_response:
                raise VehicleException('request refused')

            data.append(response.data)

        return data

    def unlock(self, key: bytes | None = None):
        '''mode $27 - security access mode'''
        if key is None:
            assert self.pcm_type is not None

            seed_request = VpwMessage(
                Priority.physical0,
                PhysicalAddress.pcm,
                PhysicalAddress.scantool,
                Mode.unlock,
                0x01
            )

            seed_response = self._device.send_message(seed_request)[0]
            key = seedkey(seed_response.data, self.pcm_type.seedkey_algorithm)

        unlock_request = VpwMessage(
            Priority.physical0,
            PhysicalAddress.pcm,
            PhysicalAddress.scantool,
            Mode.unlock,
            0x02,
            key
        )

        unlock_response = self._device.send_message(unlock_request)[0]
        response_code = unlock_response.data[0]
        match response_code:
            case 0x34:
                return # key accepted
            case 0x35:
                raise UnlockException('key rejected')
            case 0x36:
                raise UnlockException('too many unlock attempts')
            case 0x37:
                raise UnlockException('time delay not expired')
            case _:
                raise UnlockException(f'unknown response code: {response_code}')

    def read_block(self, block_id: int) -> bytes:
        '''mode $3C - read data block'''
        request = VpwMessage(
            Priority.physical0,
            PhysicalAddress.pcm,
            PhysicalAddress.scantool,
            Mode.read_block,
            block_id
        )
        response = self._device.send_message(request)[0]
        return response.data
    
    def write_block(self, block_id: int, data: bytes):
        '''mode $3B - write data block'''
        request = VpwMessage(
            Priority.physical0,
            PhysicalAddress.pcm,
            PhysicalAddress.scantool,
            Mode.write_block,
            block_id,
            data
        )
        response = self._device.send_message(request)[0]

        if response.submode != request.submode:
            raise VehicleException('write failed')

    def get_vin(self) -> str:
        vin1 = self.read_block(BlockId.vin1)
        vin2 = self.read_block(BlockId.vin2)
        vin3 = self.read_block(BlockId.vin3)
        vin_bytes = bytes((*vin1[1:], *vin2, *vin3))
        return vin_bytes.decode('ASCII')

    def write_vin(self, vin: str):
        if not re.match(r'\b[(A-H|J-N|P|R-Z|0-9)]{17}\b', vin):
            raise ValueError('invalid VIN')

        vin_bytes = vin.encode('ASCII')
        self.write_block(BlockId.vin1, bytes((0x00, *vin_bytes[:5]))) # first byte is 0x00
        self.write_block(BlockId.vin2, vin_bytes[5:11])
        self.write_block(BlockId.vin3, vin_bytes[11:])

    def get_osid(self) -> int:
        return int.from_bytes(self.read_block(BlockId.osid))


from typing import Any

DPID_MAX = 0xFE
DPID_MIN = 0xF2
DPID_MAX_BYTES = 6

class Pid:
    def __init__(self, name: str, pid: int, size: int, decoder=lambda x: x):
        assert pid in range(0xFFFF)
        self.name = name
        self.id = pid
        self.size = size # number of data bytes returned
        self.decoder = decoder

    def __bytes__(self):
        return self.id.to_bytes(2)

    def __eq__(self, other):
        if isinstance(other, Pid):
            return self.id == other.id
        elif isinstance(other, int):
            return self.id == other
        elif isinstance(other, bytes):
            return bytes(self) == other
        else:
            raise NotImplementedError

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

class Dpid:
    '''diagnostic data packet'''
    def __init__(self, dpid: int):
        assert dpid in range(0xFF)
        self.id = dpid
        self.pids = []

    def __bytes__(self):
        return self.id.to_bytes()

    def __eq__(self, other):
        if isinstance(other, Dpid):
            return self.id == other.id
        elif isinstance(other, int):
            return self.id == other
        elif isinstance(other, bytes):
            return bytes(self) == other
        else:
            raise NotImplementedError

    def __hash__(self):
        return hash(self.id)

    def __contains__(self, key):
        if isinstance(key, Pid):
            return key in self.pids
        elif isinstance(key, int):
            return key in [pid.id for pid in self.pids]
        elif isinstance(key, bytes):
            return key in [bytes(pid) for pid in self.pids]
        else:
            raise NotImplementedError

    def __len__(self):
        '''return length of response data in bytes'''
        return sum([pid.size for pid in self.pids])
    
    @property
    def bytes_free(self):
        return DPID_MAX_BYTES - len(self)

    def unpack(self, data: bytes) -> dict[Pid, bytes]:
        values = dict.fromkeys(self.pids)
        read_byte = 0
        for pid in values:
            values[pid] = data[read_byte: read_byte + pid.size]
            read_byte += pid.size

        return values

class DpidLogger:
    def __init__(self, vehicle):
        self._vehicle = vehicle
        self.pids = {}
        self._dpids = []
        self._avaliable_dpids = [Dpid(i) for i in range(DPID_MIN, DPID_MAX+1)]

    def add_pid(self, pid: Pid):
        '''add PID to data logger'''
        assert pid not in self.pids
        
        for dpid in self._dpids:
            if dpid.bytes_free >= pid.size:
                self._vehicle.define_dpid(dpid.id, pid.id, pid.size, len(dpid)+1)
                dpid.pids.append(pid)
                self.pids.update({pid: dpid})
                return
        
        dpid = self._avaliable_dpids.pop()
        self._vehicle.define_dpid(dpid.id, pid.id, pid.size, 1) # offset 1 is the first data byte
        dpid.pids.append(pid)
        self.pids.update({pid: dpid})
        self._dpids.append(dpid)

    def remove_pid(self, pid: Pid):
        '''remove PID from data logger'''
        dpid = self.pids[pid]
        self._vehicle.define_dpid(dpid.id, pid.id, 0, 0) # size=0 offset=0 removes pid
        dpid.pids.remove(pid)
        self.pids.pop(pid)

    def get_row(self) -> dict[Pid, Any]:
        '''
        query vehicle for all PIDs being logged
        returns dict of PIDs and their corresponding decoded value
        '''
        values = dict.fromkeys(self.pids)
        for request_dpids in [self._dpids[i:i + 6] for i in range(0, len(self._dpids), 6)]: # request groups of <= 6 dpids
            response = self._vehicle.get_dpids([dpid.id for dpid in request_dpids])
            for i, data in enumerate(response): # one respose line per dpid
                values.update(request_dpids[i].unpack(data))

        return {pid: pid.decoder(value) for pid, value in values.items()}


import serial
from enum import IntEnum
from .vpw import VpwMessage
from .exceptions import DeviceException

import logging
logger = logging.getLogger(__name__)

ELM_PROMPT = b'>'

class Device:
    '''scantool base class'''

    def set_header(self, header: bytes):
        raise NotImplementedError('this is only implemented in derived classes')

    def send_command(self, command: str, num_lines: int | None = None) -> list[str]:
        raise NotImplementedError('this is only implemented in derived classes')

    def send_message(self, message: VpwMessage, num_lines: int | None = None) -> list[VpwMessage]:
        '''send VpwMessage and return responses'''

        if self._header != message.get_header():
            self.set_header(message.get_header())

        response = self.send_command(repr(message), num_lines)

        data_frames = []
        for line in response:
            try:
                data_frames.append(bytes.fromhex(line))
            except ValueError:
                logger.warning(f'non-hex data: {line}')

        data_start = 4 + len(message.submode)
        messages = []
        for frame in data_frames:
            try:
                response_message = VpwMessage(
                    frame[0], # priority
                    frame[1], # target address
                    frame[2], # source address
                    frame[3], # mode
                    frame[4:data_start], # submode
                    frame[data_start:-1] # data
                )

                crc = frame[-1] # checksum

            except IndexError:
                logger.warning(f'invalid frame: {frame.hex()}')

            # TODO validate checksum here, maybe check mode/submode too?
            
            if response_message.mode != message.mode + 0x40:
                logger.warning('unexpected response mode')

            messages.append(response_message)

        if len(messages) == 0:
            raise DeviceException('no valid data received')

        return messages

class ElmProtocol(IntEnum):
    '''ELM327 protocols'''
    auto = 0
    j1850vpw = 2

class Elm327(Device):
    '''handles serial communication with ELM327 scantools'''

    def __init__(self, portname: str, **kwargs):
        self._baudrate = kwargs.pop('baudrate', 115200)
        self._timeout = kwargs.pop('timeout', 1)

        self._port = serial.Serial(
            port=portname,
            baudrate=self._baudrate,
            parity=serial.PARITY_NONE,
            stopbits=1,
            bytesize=8,
            timeout=self._timeout
        )

        # initalize device
        self.send_command('AT Z') # reset
        self.send_command('AT E0') # disable echo
        self.send_command('AT S0') # disable spaces
        self.send_command('AT H1') # display headers
        self.send_command('AT AL') # allow long (>7 byte) messages

        self.set_protocol(
            kwargs.pop('protocol', ElmProtocol.j1850vpw))
        
        self._header = None # current message header

    def send_command(self, command: str, num_lines: int | None = None) -> list[str]:
        '''
        send command and wait for response(s)
        num_lines should not be used to ignore messages
        ignored messages remain in elm buffer and may be treated as responses to subsequent commands
        num_lines = None -> read unitl elm timeout
        '''
        logger.debug(f'TX: {command}')

        if num_lines:
            command = command + str(num_lines) + '\r'
        else:
            command = command + '\r'

        self._port.write(command.encode('ASCII'))

        # read from serial port until ELM_PROMPT or timeout
        buffer = self._port.read_until(ELM_PROMPT)

        if len(buffer) == 0:
            raise DeviceException('no data')

        assert buffer.endswith(ELM_PROMPT)
        buffer = buffer[:-1]

        # decode buffer, split lines, remove empty lines, remove whitespace 
        string = buffer.decode('ASCII')
        lines = [line.strip() for line in string.split('\r') if line]

        logger.debug(f'RX: {lines}')

        if '?' in lines:
            raise DeviceException('invalid message')

        if num_lines and (len(lines) != num_lines):
            raise DeviceException(f'expected {num_lines} responses but received {len(lines)}')

        return lines

    def set_header(self, header: bytes):
        '''set message header'''

        if 'OK' not in self.send_command(f'ATSH {header.hex()}'):
            raise DeviceException('set header failed')

        self._header = header

    def set_protocol(self, protocol: int):
        '''set protocol'''

        if 'OK' not in self.send_command(f'ATSP{protocol}'):
            raise DeviceException('set protocol failed')


from .utils import get_bytes
from enum import IntEnum

class Priority(IntEnum):
    '''
    First byte of 3 byte header defines message priority and type
    See SAE J2178-1 section 7.2

    bits 7,6,5:  priority (0-7)
    bit 4:       0 = 3 byte header
    bit 3:       IFR: 1=not allowed (GM), 0=required
    bit 2:       addressing mode: 1=physical, 0=functional
    bit 1,0:     message type
    '''
    functional0 = 0x68 # priority 3, GM, functional, Type: Function Command/Status
    physical0 = 0x6C # priority 3, GM, physical, Type: Node-to-Node

class DataRate(IntEnum):
    '''
    1 byte value representing how data should be returned
    See SAE J2190 section 5.10
    Applies to modes $21-$23, $2A, and probably others
    '''
    single_response = 0x01
    repeat_slow = 0x02
    repeat_medium = 0x03
    repeat_fast = 0x04
    stop_transmission = 0x00

class PhysicalAddress(IntEnum):
    scantool = 0xF0
    pcm = 0x10
    broadcast = 0xFE

class FunctionalAddress(IntEnum):
    obd_request = 0x6A
    obd_response = 0x6B

class Mode(IntEnum):
    # SAE J1979 legislated modes
    get_pid = 0x01
    get_freeze_frame = 0x02
    get_dtc = 0x03
    clear_dtc = 0x04
    get_test_results = 0x06
    get_pending_dtcs = 0x07
    get_vehicle_info = 0x09

    # SAE J2190 modes
    read_block = 0x3C
    write_block = 0x3B
    unlock = 0x27
    get_pid_ext = 0x22
    define_dpid = 0x2C
    get_dpid = 0x2A
    download_request = 0x34
    data_transfer = 0x36
    test_device_present = 0x3F
    general_response = 0x7F

class VpwMessage:
    '''SAE J1850 VPW message'''

    def __init__(self, priority, target_address, source_address, mode, submode=b'', data=b''):
        assert priority in range(0xFF)
        assert target_address in range(0xFF)
        assert source_address in range(0xFF)
        assert mode in range(0xFF)

        self.priority = priority
        self.target_address = target_address
        self.source_address = source_address
        self.mode = mode
        self.submode = get_bytes(submode) # submode is typically echoed in response
        self.data = get_bytes(data)

    def get_header(self) -> bytes:
        '''return message header'''
        return bytes((self.priority, self.target_address, self.source_address))

    def __bytes__(self):
        '''return message bytes'''
        return bytes((self.mode, *self.submode, *self.data))

    def __repr__(self):
        '''return hex string'''
        return bytes(self).hex()

    def __getitem__(self, index):
        return bytes(self)[index]

    def __eq__(self, other):
        return (self.get_header(), bytes(self)) == (other.get_header(), bytes(other))


def aem30_0300(data: bytes):
    '''PID $114B (EGR sensor)'''
    n = int.from_bytes(data)
    v = (n * 255) / 5
    afr = (2.375 * v) + 7.3125
    lam = (0.1621 * v) + 0.4990 # lambda
    return afr

def rpm(data):
    '''PID $000C'''
    return int.from_bytes(data) * 0.25

def ect_c(data):
    '''PID $0005'''
    return int.from_bytes(data) - 40

def timing_deg(data):
    '''PID $000E'''
    return (int.from_bytes(data) / 2) - 64

def map_kpa(data):
    '''PID S000B'''
    return int.from_bytes(data)

def maf_hz(data):
    '''PID S1250'''
    return int.from_bytes(data) * 2.048

def tps(data):
    '''PID $0011'''
    return int.from_bytes(data) / 2.56

def kph(data):
    '''PID $000D'''
    return int.from_bytes(data)

def fuel_trim(data):
    '''
    LTFT bank 1 = $0007
    LTFT bank 2 = $0009
    STFT bank 1 = $0006
    STFT bank 2 = $0008
    '''
    return (int.from_bytes(data) - 128) / 1.28